import pathlib
import pandas as pd
import numpy as np
import scipy.stats


def load_data2(data_path):
    """
    Load new data from 2020-12-23. Note that some variables were added that might need to excluded from training.
    This also includes microbiome.
    """

    data_path = pathlib.Path(data_path) / "default/milk/2020-12-23/IMiC Data"

    data_milk = pd.read_excel(
        data_path / "20201218_IMiC_Data_Update_02_no-password.xlsx")

    # normalizing NaNs

    data_milk[data_milk == 8888] = np.nan
    data_milk[data_milk == 888] = np.nan
    data_milk[data_milk == 999] = np.nan

    # for infant growth variables birth ... not really relevant
    data_milk[data_milk == 99.9] = np.nan

    # this is problematic since sometime 8 might mean something ... needs to be done on a per column basis
    # data_milk[data_milk == 8] = np.nan
    # this should do it
    # TODO: check!
    print("Replace 8 with NaN for some binary variables")
    for c, n_unique in zip(data_milk.columns, data_milk.nunique()):
        if n_unique <= 4:
            unique = data_milk[c].unique()
            if 8 in unique:
                print("*", c)
                data_milk.loc[data_milk[c].values == 8, c] = np.nan

    # categorize variables

    data_milk_variables = pd.read_excel(
        data_path / "20201218_IMiC_Data_Update_02_no-password.xlsx",
        sheet_name=2)

    index_tuples = data_milk_variables.values[:, :4]
    last_c1 = None
    last_c2 = None
    column_groups = []
    for c1, c2, c3, c4 in index_tuples:
        #     print(c1,c2)

        if type(c1) == float:  # wtf
            c1 = last_c1
        elif type(c1) == str and c1.startswith("HMOs"):
            c1 = last_c1
        else:
            c1 = c1.strip()
            last_c1 = c1
            last_c2 = "general"

        if type(c2) == float:  # wtf
            c2 = last_c2
        else:
            c2 = c2.strip()
            last_c2 = c2

        # microbiome data is missing from "Data" sheet
        if c2 != "Microbiota":
            column_groups.append((c1, c2, c4))

    # print(len(data_milk.columns), len(column_groups))
    assert len(data_milk.columns) == len(column_groups)

    # check variable names between data to variable sheet
    for dc, c in zip(data_milk.columns, column_groups):
        assert(dc in c[2])

    data_milk.columns = pd.MultiIndex.from_tuples([
        (*cg[:2], c) for cg, c in zip(column_groups, data_milk.columns)])

    # some asserts to make sure we didn't mess up
    # TODO: maybe check
    assert data_milk.columns[3 - 2][1] == "Study Site"
    assert data_milk.columns[4 - 2][1] == "Number of people in house"

    data_milk.set_index(("Identify variables", "CHILDid", "CHILDid"), inplace=True)
    data_milk.index.name = "CHILDid"

    # add microbiome

    # family aggregation
    data_family = pd.read_excel(data_path / "Milk_Family_Table_01122020_no-password.xlsx", index_col="CHILDid")
    data_family.columns = pd.MultiIndex.from_tuples(
        [("Breast milk data", "Microbiome_Family", f"{i:03d}___{c}") for i, c in enumerate(data_family.columns)])

    # genus aggregation
    data_genus = pd.read_excel(data_path / "Milk_Genus_Table_01122020_no-password.xlsx", index_col="CHILDid")
    data_genus.columns = pd.MultiIndex.from_tuples([
        ("Breast milk data", "Microbiome_Genus", f"{i:03d}___{c}") for i, c in enumerate(data_genus.columns)])

    # all microbiome data
    data_asv = pd.read_excel(data_path / "Milk_ASV_Table_01122020_no-password.xlsx", index_col="CHILDid")
    data_asv_taxa = pd.read_excel(data_path / "Milk_ASV_Taxa_Table_01122020_no-password.xlsx", index_col=0)

    data_asv_columns_map = {}
    for i, r in data_asv_taxa.iterrows():
        column = []
        for n, v in zip(r.index, r.values):
            column.append(f"{n}-{v}")
        data_asv_columns_map[i] = "___".join(column)

    data_asv.columns = pd.MultiIndex.from_tuples([
        ("Breast milk data", "Microbiome_ASV", data_asv_columns_map[c]) for c in data_asv.columns])

    data_milk = data_milk.join([data_family, data_genus, data_asv])

    # add bmi growth variables

    bmi_3m = (data_milk["Infant Growth"]["Weight"]["CHLTH3MQ1"] / 1000) / (
                data_milk["Infant Growth"]["Length"]["CHLTH3MQ2"] / 100) ** 2
    bmi_1y = (data_milk["Infant Growth"]["Weight"]["CHCLA1YQ1"] / 1000) / (
                data_milk["Infant Growth"]["Length"]["CHCLA1YQ2"] / 100) ** 2

    zbmi_3m = scipy.stats.zscore(bmi_3m, nan_policy="omit")
    zbmi_1y = scipy.stats.zscore(bmi_1y, nan_policy="omit")

    data_milk[("Infant Growth", "BMI", "bmi_3m")] = bmi_3m
    data_milk[("Infant Growth", "BMI", "bmi_1y")] = bmi_1y
    data_milk[("Infant Growth", "BMI", "zbmi_3m")] = zbmi_3m
    data_milk[("Infant Growth", "BMI", "zbmi_1y")] = zbmi_1y

    return data_milk


def load_data(data_path):
    """
    Load original data from July 2020.
    """

    data_path = pathlib.Path(data_path) / "default/milk/2020-07-16"

    data_milk = pd.read_excel(data_path / "IMiC_Data_milk-07162020_no-password.xlsx")

    # normalizing NaNs

    data_milk[data_milk == 8888] = np.nan
    data_milk[data_milk == 888] = np.nan
    data_milk[data_milk == 999] = np.nan

    # this is problematic since sometime 8 might mean something ... needs to be done on a per column basis
    # data_milk[data_milk == 8] = np.nan
    # this should do it
    # TODO: check!
    print("Replace 8 with NaN for some binary variables")
    for c, n_unique in zip(data_milk.columns, data_milk.nunique()):
        if n_unique <= 4:
            unique = data_milk[c].unique()
            if 8 in unique:
                print("*", c)
                data_milk.loc[data_milk[c].values == 8, c] = np.nan

    # categorize variables

    data_milk_variables = pd.read_excel(
        data_path / "IMiC_Data_milk-07162020_no-password.xlsx",
        sheet_name=2)

    index_tuples = data_milk_variables.values[:, :2]
    last_c1 = None
    last_c2 = None
    column_groups = []
    for c1, c2 in index_tuples:
        #     print(c1,c2)

        if type(c1) == float:  # wtf
            c1 = last_c1
        elif type(c1) == str and c1.startswith("HMOs"):
            c1 = last_c1
        else:
            c1 = c1.strip()
            last_c1 = c1
            last_c2 = "general"

        if type(c2) == float:  # wtf
            c2 = last_c2
        else:
            c2 = c2.strip()
            last_c2 = c2

        # microbiome data is missing from "Data" sheet
        if c2 != "Microbiota":
            column_groups.append((c1, c2))

    # print(len(data_milk.columns), len(column_groups))
    assert len(data_milk.columns) == len(column_groups)

    data_milk.columns = pd.MultiIndex.from_tuples([(*cg, c) for cg, c in zip(column_groups, data_milk.columns)])

    # some asserts to make sure we didn't mess up

    assert data_milk.columns[3 - 2][1] == "Study Site"
    assert data_milk.columns[4 - 2][1] == "Number of people in house"

    assert data_milk.columns[55 - 2][1] == "Animals (pets)"
    assert data_milk.columns[56 - 2][1] == "Age"

    assert data_milk.columns[482 - 2][1] == "Growth Trajectories"
    assert data_milk.columns[483 - 2][1] == "Fever"

    assert data_milk.columns[677 - 2][1] == "Hormones"
    assert data_milk.columns[678 - 2][1] == "Cytokines"

    data_milk.set_index(("Identify variables", "CHILDid", "CHILDid"), inplace=True)
    data_milk.index.name = "CHILDid"

    return data_milk