import numpy as np
import sklearn.base
import sklearn.utils


class FeatureRegressor(sklearn.base.BaseEstimator, sklearn.base.RegressorMixin):

    def __init__(self, feature_function="mean"):
        self.feature_function = feature_function

    def fit(self, X, y=None, groups=None, **kwargs):
        return self

    def predict(self, X):
        if isinstance(self.feature_function, str):
            if self.feature_function == "mean":
                return np.mean(X, axis=tuple(np.arange(1, X.shape.size)))
            elif self.feature_function == "median":
                return np.median(X, axis=tuple(np.arange(1, X.shape.size)))
            else:
                raise Exception(f"Unknown feature function: {self.feature_function}")
        elif isinstance(self.feature_function, int):
            return X[:, self.feature_function].flatten()
        else:
            # noinspection PyCallingNonCallable
            return self.feature_function(X)


class BaseClassifier(sklearn.base.BaseEstimator, sklearn.base.ClassifierMixin):

    def __init__(self):
        self.label_encoder = sklearn.preprocessing.LabelEncoder()

    def fit(self, X, y):
        X, y = sklearn.utils.check_X_y(X, y)
        y = self.label_encoder.fit_transform(y)
        return X, y

    def predict(self, X):
        X = sklearn.utils.check_array(X)
        # noinspection PyUnresolvedReferences
        probs = self.predict_proba(X)
        idx = np.argmax(probs, axis=1)
        return self.label_encoder.inverse_transform(idx)

    @property
    def classes_(self):
        return self.label_encoder.classes_


class FeatureClassifier(BaseClassifier):

    def __init__(self, feature_function="1d", reverse=False, postprocessing=None):
        super().__init__()
        self.feature_function = feature_function
        self.reverse = reverse
        self.postprocessing = postprocessing

    def fit(self, X, y=None, groups=None, **kwargs):
        super().fit(X, y)
        return self

    def predict_proba(self, X):
        X = sklearn.utils.check_array(X)

        if isinstance(self.feature_function, str):
            if self.feature_function == "1d":
                if X.shape[1] > 1:
                    raise ValueError(f"X should be 1-dimensional, but there are more than one feature: {X.shape[1]}")
                else:
                    values = X.flatten()
            elif self.feature_function == "mean":
                values = np.mean(X, axis=tuple(np.arange(1, len(X.shape))))
            elif self.feature_function == "median":
                values = np.median(X, axis=tuple(np.arange(1, len(X.shape))))
            else:
                raise Exception(f"Unknown feature function: {self.feature_function}")
        elif isinstance(self.feature_function, int):
            values = X[:, self.feature_function].flatten()
        else:
            # noinspection PyCallingNonCallable
            values = self.feature_function(X)

        values = np.array([-values, values]).T

        if self.reverse:
            values = -values

        if isinstance(self.postprocessing, str):
            if self.postprocessing == "prob":
                values = sklearn.preprocessing.MinMaxScaler().fit_transform(values[:, [1]])
                values = np.array([1 - values, values]).T
            else:
                raise Exception(f"Unknown postprocessing type: {self.postprocessing}")
        elif self.postprocessing is not None:
            values = self.postprocessing(values)

        return values


class BackwardSelectionEstimator(sklearn.base.BaseEstimator):
    """
    TODO: We could add a CV version of this
    """

    def __init__(self, estimator, step_size=1, depth=-1, scoring=None):
        self.estimator = estimator
        self.step_size = step_size
        self.depth = depth
        self.scoring = scoring

    def _fit_and_score(self, X, y, groups, indicator=None):
        """Fit and score estimator"""

        if indicator is not None:
            column_transformer = sklearn.compose.ColumnTransformer(
                [("select", "passthrough", indicator)],
                remainder="drop")
            X = column_transformer.fit_transform(X, y)
        else:
            column_transformer = None

        estimator = sklearn.base.clone(self.estimator)
        try:
            estimator.fit(X, y, groups=groups)
        except AttributeError:
            estimator.fit(X, y)

        scorer = sklearn.metrics.check_scoring(estimator, self.scoring)
        return column_transformer, estimator, scorer(estimator, X, y)

    # noinspection PyAttributeOutsideInit
    def fit(self, X, y, groups=None):

        # init stats
        stats = [dict(
            level=0,
            indicator=np.ones(X.shape[1], dtype=bool),
            score=self._fit_and_score(X, y, groups)[2],
            dropped=None,
            stats=None)]

        for level in range(1, X.shape[1] + 1):
            level_indicator = stats[-1]["indicator"]
            feature_info = []
            for feature_idx in np.argwhere(level_indicator):
                feature_indicator = np.copy(level_indicator)
                feature_indicator[feature_idx] = False
                _, _, score = self._fit_and_score(X, y, groups=groups, indicator=feature_indicator)
                feature_info.append((feature_idx, score))

            # select variables to drop
            feature_scores_order = np.argsort([s for _, s in feature_info])
            todrop_feature_info = feature_info[feature_scores_order[:-min(self.step_size, len(feature_info))]]
            todrop_feature_indices = [idx for idx, _ in todrop_feature_info]
            todrop_feature_scores = [score for _, score in todrop_feature_info]

            # derive new indicator vector
            new_indicator = np.copy(level_indicator)
            new_indicator[todrop_feature_indices] = False

            # add stats
            stats.append(dict(
                level=level,
                indicator=new_indicator,
                score=todrop_feature_scores[0] if len(todrop_feature_scores) == 0
                      else self._fit_and_score(X, y, groups, new_indicator),
                dropped=todrop_feature_indices,
                stats=feature_info))

            # break if there are no variables left
            if sum(new_indicator) == 0:
                break

        # select final model and train estimator
        # TODO: for this we could to cross validation
        # TODO: maybe make this selection procedure modular?
        scores = [s["score"] for s in stats]
        self.best_model_idx_ = np.argmax(scores).max()  # select the model with the least variables
        self.stats_ = stats
        self.feature_indicator_ = stats[self.best_model_idx_]["indicator"]
        self.column_transformer_ = sklearn.compose.ColumnTransformer(
            [("select", "passthrough", self.feature_indicator_)],
            remainder="drop")
        self.estimator_fit_ = sklearn.base.clone(self.estimator)
        self.estimator_fit_.fit(self.column_transformer_.fit_transform(X, y), y)

        return self

    def predict_proba(self, X):
        self.estimator_fit_.predict_proba(self.column_transformer_.transform(X))

    def predict(self, X):
        self.estimator_fit_.predict(self.column_transformer_.transform(X))

    def decision_function(self, X):
        self.estimator_fit_.decision_function(self.column_transformer_.transform(X))


class OneStepBackwardSelectionEstimator(sklearn.base.BaseEstimator):
    """
    Selects a fixed number k variables OR selects only variables that
    push the estimator below the performance of the performance of the overall estimator.
    """
    def __init__(self):
        raise NotImplementedError()


class ForwardSelectionEstimator(sklearn.base.BaseEstimator):
    """Add variables step by step. Select the best (?) model."""
    def __init__(self):
        raise NotImplementedError()


class BackwardForwardSelectionEstimator(sklearn.base.BaseEstimator):
    """
    Remove best working variables one at a time.
    And then train models for each step. After that, select best (?) model.
    """
    def __init__(self):
        raise NotImplementedError()
